#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/power.h>
#include <stdlib.h>
#include <math.h>
#include "hd44780.h"

//Red OC0A PB6
#define led_red_on()  TCCR0A |= _BV(COM0A1); PORTD |= _BV(6)
#define led_red_off() TCCR0A &= ~_BV(COM0A1); PORTD &= ~_BV(6)
//Green OC2A PB3
#define led_green_on()  TCCR2A |= _BV(COM2A1); PORTB |= _BV(3)
#define led_green_off() TCCR2A &= ~_BV(COM2A1); PORTB &= ~_BV(3)
//Blue OC2B PD3
#define led_blue_on()  TCCR2A |= _BV(COM2B1); PORTD |= _BV(3)
#define led_blue_off()  TCCR2A &= ~_BV(COM2B1); PORTD &= ~_BV(3)
//PB2
#define pb2_on()  PORTB |= _BV(2)
#define pb2_off() PORTB &= ~_BV(2)



#define MAX(x, y) (((x) > (y)) ? (x) : (y))
#define MIN(x, y) (((x) < (y)) ? (x) : (y))

#define MODE_HSB 0
#define MODE_CYCLE1 1
#define MODE_CYCLE2 2
#define MODE_PULSE1 3
#define MODE_PULSE2 4
#define MODE_STROBE 5
char g_mode=MODE_HSB;

//ADC for Hue, Sat, Bri
volatile uint16_t adc[]={0,0,0};
uint8_t  requested_adc_channel=0;
//for normalisation of ADC results
#define ADC0_LOW 16
#define ADC0_HIGH 1023
#define ADC1_LOW 16
#define ADC1_HIGH 1023
#define ADC2_LOW 16
#define ADC2_HIGH 1023

typedef struct
{
    float red;          //0 to 1
    float green;        //0 to 1
    float blue;         //0 to 1
} SColorRGB;
 
typedef struct
{
    float hue;          //0 to 359
    float saturation;   //0 to 1
    float brightness;   //0 to 1
}  SColorHSB;

SColorHSB g_hsb;
SColorRGB g_rgb;

//#define WHITE_FULL 255
#define RED_FULL   255
#define GREEN_FULL 180
#define BLUE_FULL  105

#define TTL_MODE_CHANGE 48  //units of 50ms
volatile unsigned char g_remaining_ttl_mode_change=0;

#define PULSE_UP 0
#define PULSE_TOP 1
#define PULSE_DOWN 2
#define PULSE_BOTTOM 3
volatile unsigned char g_pulse_dir=PULSE_UP;
volatile unsigned short g_remaining_ttl_pulse=0;

#define STROBE_ON 1
#define STROBE_OFF 0
#define TTL_STROBE 3
unsigned char g_strobe_state=STROBE_OFF;
unsigned short g_remaining_ttl_strobe=0;

#define BUTTON_UNBOUNCE_TTL 3
//volatile unsigned char g_remaining_button_unbounce=0;
//volatile unsigned char g_count_mode=0;
volatile unsigned char g_btn_mode=0;

volatile unsigned char g_display_refresh=1;

char g_buffer[16];

//prototypes
int main (void);
void update_display(void);
void update_led_pwm(void);
void rgbFromHSB(SColorHSB __hsb,SColorRGB * __rgb);
void main_init(void);
void adc_init(void);
void pwm_init(void);
void ReadADC_Request(uint8_t __channel);

//Functions

int main (void)
{
	
	_delay_ms(200);
	
	double_blink();
	
	main_init();
	adc_init();
	pwm_init();
	lcd_init();

	lcd_goto(0);
	lcd_puts("Welcome          ");    
	_delay_ms(750);

	g_mode=MODE_HSB;
	power_adc_enable();
	power_timer0_enable();
	power_timer1_enable();
	power_timer2_enable();
	
	lcd_init();
    
    lcd_clrscr();
	update_display();

	unsigned short i=0;
	
	while (1)
	{
		i++;
		if (i>254)
		{
			i=0;
		}

		ReadADC_Request((uint8_t)(i % 3));
		if (g_display_refresh==1)
		{
			update_display();
			g_display_refresh=0;
		}
		update_led_pwm();

		//hack to make strobe pulses only last 10ms
		if (g_mode==MODE_STROBE && g_hsb.brightness>0)
		{
			_delay_ms(10);
			g_hsb.brightness=0;
		}
		else
		{
			_delay_ms(5);
		}
		
		_delay_ms(1);
	}

    return(0);
}

/************************************************** Update LCD/LED **************************************************/

#define UD_LINE1 1
#define UD_LINE2 2
#define UD_BOTH 3

void update_display(void)
{
	lcd_goto(0);
	
	if (g_remaining_ttl_mode_change>0)
	{
		switch (g_mode)
		{
			case MODE_HSB:
				lcd_puts("Manual          ");    
				break;
			case MODE_CYCLE1:
				lcd_puts("Slow Cycle      ");    
				break;
			case MODE_CYCLE2:
				lcd_puts("Fast Cycle      ");    
				break;
			case MODE_PULSE1:
				lcd_puts("Slow Pulse      ");    
				break;
			case MODE_PULSE2:
				lcd_puts("Fast Pulse      ");    
				break;
			case MODE_STROBE:
				lcd_puts("Strobe          ");    
				break;
		}
		//sprintf(g_buffer,"            %4u", g_mode_count);
		//lcd_goto(40);                    //Position 40 is the start of line 2
		//lcd_puts(g_buffer);
	}
	else
	{
		switch (g_mode)
		{
			case MODE_HSB:
				lcd_puts("Hue   Sat   Bri ");
				sprintf(g_buffer,"%3u  %3u%%  %3u%%", (unsigned short)floor(g_hsb.hue), (unsigned short)round(g_hsb.saturation*100), (unsigned short)round(g_hsb.brightness*100));
				//sprintf(g_buffer,"%3u  %3u%%  %3u%%", (unsigned short)floor(g_hsb.hue), (unsigned short)round(g_hsb.saturation*100), g_count_mode);
				//sprintf(g_buffer,"%u  %u  %u    ", PINB, g_mode_count, g_mode_button);
				break;
			case MODE_CYCLE1:
				lcd_puts("            Bri ");  
				sprintf(g_buffer,"            %3u%%", (unsigned short)round(g_hsb.brightness*100));
				break;
			case MODE_CYCLE2:
				lcd_puts("            Bri ");
				sprintf(g_buffer,"            %3u%%", (unsigned short)round(g_hsb.brightness*100));
				break;
			case MODE_PULSE1:
				lcd_puts("Hue   Sat       ");
				sprintf(g_buffer,"%3u  %3u%%      ", (unsigned short)floor(g_hsb.hue), (unsigned short)round(g_hsb.saturation*100));
				break;
			case MODE_PULSE2:
				lcd_puts("Hue   Sat       ");
				sprintf(g_buffer,"%3u  %3u%%      ", (unsigned short)floor(g_hsb.hue), (unsigned short)round(g_hsb.saturation*100));
				break;
			case MODE_STROBE:
				lcd_puts("Hue   Sat       ");
				sprintf(g_buffer,"%3u  %3u%%      ", (unsigned short)floor(g_hsb.hue), (unsigned short)round(g_hsb.saturation*100));    
				break;
			default:
				lcd_puts("UNKNOWN MODE    ");
				_delay_ms(80);
				break;
		}
		lcd_goto(40);                    //Position 40 is the start of line 2
		lcd_puts(g_buffer);
	}
}

char *byte_to_binary(uint8_t x)
{
    static char b[9];
    b[0] = '\0';

    int z;
    for (z = 128; z > 0; z >>= 1)
    {
        strcat(b, ((x & z) == z) ? "1" : "0");
    }

    return b;
}

void update_led_pwm(void)
{
		rgbFromHSB(g_hsb,&g_rgb);
               
		//Red (OCR0A)
		if (round((RED_FULL+1)*g_rgb.red)==0)
		{
			led_red_off();
		}
		else
		{
			OCR0A = (unsigned char)((RED_FULL+1)*g_rgb.red-1);
			led_red_on();
		}

		//Green (OCR2A)
		if (round((GREEN_FULL+1)*g_rgb.green)==0)
		{
			led_green_off();
		}
		else
		{
			OCR2A = (unsigned char)((GREEN_FULL+1)*g_rgb.green-1);
			led_green_on();
		}

		//Blue (OCR2B)
		if (round((BLUE_FULL+1)*g_rgb.blue)==0)
		{
			led_blue_off();
		}
		else
		{
			OCR2B = (unsigned char)((BLUE_FULL+1)*g_rgb.blue-1);
			led_blue_on();
		}
}

void rgbFromHSB(SColorHSB __hsb,SColorRGB * __rgb)
{
        if (__hsb.saturation==0)
        {
                __rgb->red = __hsb.brightness;
                __rgb->green = __hsb.brightness;
                __rgb->blue = __hsb.brightness;
        }
        else
        {
            float max = __hsb.brightness;
			float dif = __hsb.brightness * __hsb.saturation;
			float min = __hsb.brightness - dif;
 
			if (__hsb.hue < 60)
			{
				__rgb->red = max;
				__rgb->green = __hsb.hue * dif / 60 + min;
				__rgb->blue = min;
			}
			else if (__hsb.hue < 120)
			{
				__rgb->red = -(__hsb.hue - 120) * dif / 60 + min;
				__rgb->green = max;
				__rgb->blue = min;
			}
			else if (__hsb.hue < 180)
			{
				__rgb->red = min;
				__rgb->green = max;
				__rgb->blue = (__hsb.hue - 120) * dif / 60 + min;
			}
			else if (__hsb.hue < 240)
			{
				__rgb->red = min;
				__rgb->green = -(__hsb.hue - 240) * dif / 60 + min;
				__rgb->blue = max;
			}
			else if (__hsb.hue < 300)
			{
				__rgb->red = (__hsb.hue - 240) * dif / 60 + min;
				__rgb->green = min;
				__rgb->blue = max;
			}
			else if (__hsb.hue <= 360)
			{
				__rgb->red = max;
				__rgb->green = min;
				__rgb->blue = -(__hsb.hue - 360) * dif / 60 + min;
			}
			else
			{
				__rgb->red = 0;
				__rgb->green = 0;
				__rgb->blue  = 0;
			}
	    }
	}


/************************************************** Initialisations **************************************************/

void main_init(void)
{
	//set_sleep_mode(SLEEP_MODE_IDLE);

	power_spi_disable();
	power_usart0_disable();
	
	DDRB  = 0b11111011; //1=output, 0=input
    DDRC  = 0b01110000;
    DDRD  = 0b11111111;

	PORTB = 0b00000100;
	PORTC = 0b00111111;
	PORTD = 0b00000000;

    g_hsb.hue = 0;
    g_hsb.saturation = 1;
    g_hsb.brightness = .5;

	//PCICR = _BV(PCIE0);     //Enable pin change interrupts on PB2
	//PCMSK0 = _BV(PCINT2);

	//Setup 16 bit timer
	TIMSK1 = _BV(OCIE1A);                         // Enable Interrupt Timer/Counter1, Output Compare A (TIMER1_COMPA_vect)
    TCCR1B = _BV(CS12) | _BV(CS10) | _BV(WGM12);     // Clock/1024, 0.000128 seconds per tick, Mode=CTC
    OCR1A = 391;                                    // 0.000128*391 ~= 2 TIMER1_COMPA_vect will be triggered every 50ms

	sei();  //Enable interrupts
}

void adc_init(void)
{
        //Enable ADC, set 128 prescale and enable the interrupt
        ADCSRA = _BV(ADEN) | _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0) | _BV(ADIE);

}

void pwm_init(void)
{
	//TCCR0A = _BV(COM0A1) | _BV(COM0B1) | _BV(WGM01) | _BV(WGM00);      // Non inverting mode on OC0A and OC0B, Mode = Mode 3 FAST PWM
	TCCR0A = _BV(COM0A1) | _BV(WGM01) | _BV(WGM00);
    TCCR0B = _BV(CS00);                                                // No prescaling
 
    TCCR2A = _BV(COM2A1) | _BV(COM2B1) | _BV(WGM21) | _BV(WGM20);      // Non inverting mode on OC2A and OC2B, Mode = Mode 3 FAST PWM
    TCCR2B = _BV(CS20);                                                // No prescaling
}

void double_blink(void)
{
	DDRB  = 0b11111111;
	pb2_off();
	_delay_ms(60);
	pb2_on();
	_delay_ms(60);
	pb2_off();
	_delay_ms(60);
	pb2_on();
}


/************************************************** Interrupts **************************************************/

//50ms timer
ISR(TIMER1_COMPA_vect)
{
	if(!bit_is_set(PINB,2))
	{
		//g_count_mode++;
		if (g_btn_mode==0)
		{
			g_btn_mode=1;
			//g_count_mode++;
			if (g_remaining_ttl_mode_change>0) //Already in mode change
			{
				if (g_mode==MODE_STROBE)
				{
					g_mode=MODE_HSB;
				}
				else
				{
					g_mode++;
				}
			}
			g_remaining_ttl_mode_change=TTL_MODE_CHANGE;
			g_display_refresh=1;
			power_adc_enable();
		}
	}
	else
	{
		g_btn_mode=0;
	}

	if (g_remaining_ttl_mode_change>0)
	{
		g_remaining_ttl_mode_change--;
	}
	else
	{
		if (g_remaining_ttl_mode_change==0)
		{
			g_display_refresh=1;
		}

		switch (g_mode)
		{
			case MODE_CYCLE1:
				g_hsb.hue += 0.33;
				if (g_hsb.hue>=360)
					g_hsb.hue=0;
				g_hsb.saturation=1;
				update_led_pwm();
				break;
			case MODE_CYCLE2:
				g_hsb.hue += 4;
				if (g_hsb.hue>=360)
					g_hsb.hue=0;
				g_hsb.saturation=1;
				update_led_pwm();
				break;
			case MODE_PULSE1:
				if (g_pulse_dir==PULSE_UP)
				{
					g_hsb.brightness += 0.04;
					if (g_hsb.brightness>=1)
					{
						g_hsb.brightness=1;
						g_pulse_dir=PULSE_TOP;
						g_remaining_ttl_pulse=8;
					}
				}
				else if (g_pulse_dir==PULSE_TOP)
				{
					g_remaining_ttl_pulse--;
					if (g_remaining_ttl_pulse==0)
					{
						g_pulse_dir=PULSE_DOWN;
					}
				}
				else if (g_pulse_dir==PULSE_DOWN)
				{
					g_hsb.brightness -= 0.04;
					if (g_hsb.brightness<=0.2)
					{
						//g_hsb.brightness=0;
						g_pulse_dir=PULSE_BOTTOM;
						g_remaining_ttl_pulse=8;
					}
				}
				else if (g_pulse_dir==PULSE_BOTTOM)
				{
					g_remaining_ttl_pulse--;
					if (g_remaining_ttl_pulse==0)
					{
						g_pulse_dir=PULSE_UP;
					}
				}
				break;
			case MODE_PULSE2:
				if (g_pulse_dir==PULSE_UP)
				{
					g_hsb.brightness += 0.2;
					if (g_hsb.brightness>=1)
					{
						g_hsb.brightness=1;
						g_pulse_dir=PULSE_TOP;
						g_remaining_ttl_pulse=3;
					}
				}
				else if (g_pulse_dir==PULSE_TOP)
				{
					g_remaining_ttl_pulse--;
					if (g_remaining_ttl_pulse==0)
					{
						g_pulse_dir=PULSE_DOWN;
					}
				}
				else if (g_pulse_dir==PULSE_DOWN)
				{
					g_hsb.brightness -= 0.2;
					if (g_hsb.brightness<=0.3)
					{
						//g_hsb.brightness=0;
						g_pulse_dir=PULSE_BOTTOM;
						g_remaining_ttl_pulse=3;
					}
				}
				else if (g_pulse_dir==PULSE_BOTTOM)
				{
					g_remaining_ttl_pulse--;
					if (g_remaining_ttl_pulse==0)
					{
						g_pulse_dir=PULSE_UP;
					}
				}
				break;
			case MODE_STROBE:
				if (g_remaining_ttl_strobe>0)
					g_remaining_ttl_strobe--;
				else if (g_strobe_state==STROBE_ON)
				{
					g_strobe_state = STROBE_OFF;
					g_remaining_ttl_strobe =  TTL_STROBE;
					g_hsb.brightness=0;
				}
				else
				{
					g_strobe_state = STROBE_ON;
					g_hsb.brightness=1;
				}
				break;
		}
	}


}

// 3 pots (Hue, Saturation, Brightness)
ISR(ADC_vect,ISR_NOBLOCK)
{        
		adc[requested_adc_channel]=ADC;    // Store the resulting value to the adc results array
		float adc_val = (float)ADC;
		
		float hue=0;
		float sat=0;
		float bri=0;

		switch(requested_adc_channel)
		{
			//Hue
			case 0:
				if (g_mode!=MODE_CYCLE1 && g_mode!=MODE_CYCLE2)
				{
					hue = MIN(359.9,((MIN(MAX(ADC0_LOW,adc_val),ADC0_HIGH)-ADC0_LOW)*360)/(ADC0_HIGH-ADC0_LOW));
					g_hsb.hue=hue;
					//g_hsb.hue=adc_val;
				}
				break;
			//Saturation
			case 1:
				if (g_mode!=MODE_CYCLE1 && g_mode!=MODE_CYCLE2)
				{
					sat = (MIN(MAX(ADC1_LOW,adc_val),ADC1_HIGH)-ADC1_LOW)/(ADC1_HIGH-ADC1_LOW);
					g_hsb.saturation=sat;
					//g_hsb.saturation=adc_val;
				}
				break;
			//Brightness
			case 2:
				if (g_mode!=MODE_PULSE1 && g_mode!=MODE_PULSE2 && g_mode!=MODE_STROBE)
				{
					bri = (MIN(MAX(ADC2_LOW,adc_val),ADC2_HIGH)-ADC2_LOW)/(ADC2_HIGH-ADC2_LOW);
					g_hsb.brightness=bri;
					//g_hsb.brightness=adc_val;
				}
				break;
		}
		g_display_refresh=1;

}

/************************************************** AD Converter **************************************************/


void ReadADC_Request(uint8_t __channel)
{
	if (bit_is_set(ADCSRA,ADSC))
	{
			//If another ADC read is in progress, display a message and halt processing
			lcd_goto(0);
			lcd_puts("ADC Error");
			power_adc_disable();
			power_adc_enable();
			abort();
	}
	else
	{
			requested_adc_channel = __channel;        //Well need this when reading the value
   
			ADMUX = (ADMUX & 0x11110000) | __channel; //Channel selection
			ADCSRA |= _BV(ADSC);                      //Start conversion
	}
}

